#include "EMPif.h"
#include <stdio.h>
#include <string.h>
#include "util.h"
#include "portable_fileio.h"

static bool is_float(int mode) {
    switch(mode) {
    case PIF_SHORT_FLOAT:
    case PIF_SHORT_FLOAT_COMPLEX:
    case PIF_FLOAT:
    case PIF_FLOAT_INT:
    case PIF_FLOAT_COMPLEX:
    case PIF_FLOAT_INT_COMPLEX:
    case PIF_MAP_FLOAT_SHORT:
    case PIF_MAP_FLOAT_INT:
	return true;
    default:
	return false;
    }
    return false;
}

static int get_mode_size(int mode) {
    int mode_size = 0;
    
    switch(mode) {
    case PIF_CHAR:
    case PIF_BOXED_DATA:
	mode_size = sizeof(char);
	break;
    case PIF_SHORT:
    case PIF_SHORT_FLOAT:
    case PIF_SHORT_COMPLEX:
    case PIF_SHORT_FLOAT_COMPLEX:
    case PIF_MAP_FLOAT_SHORT:
	mode_size = sizeof(short);
	break;
    case PIF_FLOAT:
    case PIF_FLOAT_INT:
    case PIF_FLOAT_COMPLEX:
    case PIF_FLOAT_INT_COMPLEX:
    case PIF_MAP_FLOAT_INT:
	mode_size = sizeof(int);
	break;
    default:
	break;
    }
    return mode_size;
}

    
EMPif::EMPif(const char* filename) {
    mode_size = 0;
    right_order = true;
    valid_header = false;
    is_zipped = false;
    real_scale_factor = 1;
    
    in = open_image_file(filename, &is_zipped);
    
    if (!in) {
	printf("Error: cannot open file '%s'\n", filename);	
    }
    else {
	if (fread(&pfh, sizeof(PifFileHeader), 1, in) == 1) {
	    right_order = check_byte_order((char*)&pfh.htype);
	    
	    if (!right_order) {
		swapint((char*)&pfh.htype, 1);
		swapint((char*)&pfh.magic, 2);
	    }

	    if (pfh.magic[0] == PIF_MAGIC_NUM &&
		pfh.magic[1] == PIF_MAGIC_NUM) {
		valid_header = true;
	    }
	    
	    if (valid_header) {
		if (!right_order) {
		    swapint((char*)&pfh.mode, 1);
		    swapint((char*)&pfh.nx, 1);
		    swapint((char*)&pfh.ny, 1);
		    swapint((char*)&pfh.nz, 1);
		    swapint((char*)&pfh.nimg, 1);
		}
		
		if (is_float(pfh.mode)) {
		    real_scale_factor = atof(pfh.scalefactor);
		}
		
		mode_size = get_mode_size(pfh.mode);

		if (is_complex_mode()) {
		    pfh.nx *= 2;
		}
	    }
	}
    }
}

EMPif::~EMPif() {
    if (is_zipped) {
	pclose(in);
    }
    else {
	fclose(in);
    }
}



void EMPif::dump() const {
    printf("nx = %d, ny = %d, nz = %d, nimg = %d, ",
	   pfh.nx, pfh.ny, pfh.nz, pfh.nimg);
    
    printf("mode = %d, mode_size = %d, scale factor = %f\n",
	   pfh.mode, mode_size, real_scale_factor);
}

// return 0 if OK; 1 if bad;
int EMPif::read_data(int which_img, float** p_data, bool is_3d) {
    if (which_img < 0 || which_img > pfh.nimg || !p_data || !(*p_data)) {
	return 1;
    }
    if (!is_valid_header()) {
	return 1;
    }
    
    int pih_sz = sizeof(PifImageHeader);
    int file_offset = sizeof(PifFileHeader) +
	(pih_sz + pfh.nx * pfh.ny * mode_size) * which_img + pih_sz;

    portable_fseek(in, file_offset, SEEK_SET);
    
    int buf_size = pfh.nx * mode_size;
    unsigned char* buf = (unsigned char*) malloc(buf_size);    
    int buf_data_size = sizeof(unsigned char) * 8;
    int num_layers = 0;
    
    if (is_3d) {
	num_layers = pfh.nz;
    }
    else {
	num_layers = 1;
    }

    for (int l = 0; l < num_layers; l++) {
	
	int offset1 = l * pfh.nx * pfh.ny; 
	
	for (int j = 0; j < pfh.ny; j++) {
	    if (fread(buf, mode_size, pfh.nx, in) != pfh.nx) {
		free(buf);
		buf = 0;
		return 1;
	    }
	    
	    if (!right_order) {
		if (mode_size == sizeof(short)) {
		    swapshort((char*)buf, pfh.nx);
		}
		else if (mode_size == sizeof(int)) {
		    swapint((char*)buf, pfh.nx);
		}
	    }
	    
	    int offset2 = offset1 + j * pfh.nx;
	    
	    for (int k = 0; k < pfh.nx; k++) {
		float curr_data = 0;
		
		if (mode_size == sizeof(char)) {
		    curr_data = (float) buf[k];
		}
		else if (mode_size == sizeof(short)) {
		    curr_data = (float) ((short*)buf)[k];
		}
		else if (mode_size == sizeof(int)) {
		    curr_data = (float) ((int*)buf)[k];
		}
		(*p_data)[offset2 + k] = curr_data * real_scale_factor;
	    }
	}
    }
    
    free(buf);
    buf = 0;
    return  0;
}




