// geodcmp  6/24/02  Steve Ludtke
// Copyright 2002  Baylor College of Medicine

//N geocmp
//F This will determine a set of orthogonal basis images from a set of aligned particles
//T LA
//1
//P <images>	Input file, images to be analyzed
//P <basis>	Output file, set of orthogonal basis images
//P [n=<n basis>]	Number of basis images to output (<= number of input images)
//P [mask=<radius>]	Mask radius in pixels (use the smallest reasonable value)
//U geodcmp proj.hed basis.hed
//D This program is designed to produce a set of orthogonal basis images to
//E use for noise reduction on raw particles. Typically this program will
//E be used on a set of projections covering the asymmetric triangle, but
//E with no in-plane rotation. The resulting basis images can then be applied
//E to an (aligned) noisy image to act as a sort of filter. 

#include <stdio.h>
#include <math.h>
#include "util.h"
#include "EMData.h"

//#define N 200
//#define M 20000

main(int argc,char *argv[]) {
int i,j,k;
int mask=-1,nbasis=-1,imsize=-1,jacobi=0,quiet=0,weight=0;
EMData data;

appinit(argc,argv);

if (argc<3) appusage(argv[0]);
for (i=3; i<argc; i++) {        
	if (strncmp(argv[i],"mask=",5)==0) mask=atoi(&argv[i][5]);
	else if (strncmp(argv[i],"n=",2)==0) nbasis=atoi(&argv[i][2]);
	else if (strncmp(argv[i],"altwt",5)==0) weight=1;
	else if (strncmp(argv[i],"usefs",5)==0) ;
	else if (strncmp(argv[i],"quiet",5)==0) quiet=1;
	else { printf("Unknown parameter '%s'\n",argv[i]); exit(1); }
}

int nimg=fileCount(argv[1]);
if (nimg==0) error(ERR_EXIT,"No readable images in input file");
if (nbasis<0) nbasis=nimg;

int Ref;
if (!quiet) {
	Ref=LOGbegin(argc,argv);
	LOG(Ref,argv[1],LOG_INFILE,NULL);
}

float *eul[3];
int *euln = (int *)malloc(nimg*sizeof(float));
eul[0]=(float *)malloc(nimg*sizeof(float));
eul[1]=(float *)malloc(nimg*sizeof(float));
eul[2]=(float *)malloc(nimg*sizeof(float));

EMData **vec = (EMData **)malloc(nbasis*sizeof(EMData *));

// Read the Eulers so we can determine ordering
// the first vector is the first image
for (i=0; i<nimg; i++) {
	EMData tmp;
	tmp.readImage(argv[1],i);
	if (i==0) vec[0]=tmp.copy(0,0);
//	else vec[0]->add(&tmp);
	eul[0][i]=tmp.alt();
	eul[1][i]=tmp.az();
	eul[2][i]=tmp.phi();
	euln[i]=i;
}

printf("0\t0\t%1.5g\n",vec[0]->Sigma());
vec[0]->multConst(1.0/sqrt(vec[0]->SumSq()));

// Now create the rest of the vectors
for (i=1; i<nbasis; i++) {
	// Find the next appropriate vector
	// very much like a simple shell sort
	float bd=0;
	int bj=0;
	// We're looking for the point as far as possible from neighbors
	for (j=i; j<nimg; j++) {
		float d=1.0e30;
		for (k=0; k<i; k++) {
			float f=hypot(eul[0][k]-eul[0][j],eul[1][k]-eul[1][j]);
			if (f<d) d=f;
		}
		if (d>bd) { bd=d; bj=j; }
	}
	
	// swap the best one we found with the current one
	if (bj!=i) {
		float f;
		f=eul[0][bj]; eul[0][bj]=eul[0][i]; eul[0][i]=f;
		f=eul[1][bj]; eul[1][bj]=eul[1][i]; eul[1][i]=f;
		f=eul[2][bj]; eul[2][bj]=eul[2][i]; eul[2][i]=f;
		int I;
		I=euln[bj]; euln[bj]=euln[i]; euln[i]=I;
	}		

	// now generate the next orthogonal vector
	EMData *tmp = new EMData;
	tmp->readImage(argv[1],euln[i]);
	
	for (j=0; j<i; j++) {
		float d=tmp->dot(vec[j]);
		
		EMData *cp=vec[j]->copy(0,0);
		cp->multConst(d);
		tmp->subtract(cp);
		delete cp;
	}
//	printf("%1.2f\t%1.2f\t%d\t%1.5g\n",eul[0][i]*180./PI,eul[1][i]*180./PI,euln[i],tmp->Sigma());
	printf("%d\t%1.5g\n",i,tmp->Sigma());
	tmp->multConst(1.0/sqrt(tmp->SumSq()));
	vec[i]=tmp;
}

for (i=0; i<nbasis; i++) vec[i]->writeIMAGIC(argv[2],i);

}
