// search for local symmetry axis	Wen Jiang, Jan 2001
//
//N symAxisSearch
//F This program is used to iteratively search for the local rotation symmtry axis
//T X
//1
//P <map>	input 3D map with the symmetry axis nearly along Z-axis
//P	<n>	the rotation symmetry
//P	<sym coeff file>	the list of symmetry coeffecients at different slices along the axis
//P	[outmap=<file.mrc>]	output the 3D map that trasnformed to let the symmetry axis along the Z-axis and through the map center (nx/2,ny/2,nz/2)
//P	[maskout=<output MRC file>]	after searching the symmetry axis, mask out the symmtric voxels above certain threshold 
//P	[symrad=<pixels>]	radius of the symmtric region
//P	[srchrad=<pixels>]	radius of search region
//P	[thresh=<min sym coeff>]	minimal symmetry coeff to be included in axis fitting and density masking
//P	[zslice=<z0,z1>]	slices to search
//P	[start=<x,y>]	start axis position in xy plane, in pixel 
//P	[convergat=<degrees,pixels>]	the criterion to stop the iterative search, default to 1 degree and 1 pixel
//P	[cpu=<n>]	number of CPUs to use. parallelized with OpenMP 
//P	[verbose]	if set, intermediate results are outputed
//U symAxisSearchOMP map.mrc 3 sym-coeff.dat cpu=4 start=60,66 thresh=0.8 zslice=30,92 maskout=map-masked.mrc outmap=map-sym.mrc
//D This program will take the input 3D map, assuming the rotation symmetry axis is near the Z axis, find the best
//E	symmetry center in each slice, least-square-fit to a straight line to determine the direction and position of the  
//E symmetry axis. 
//D The parallelization is done with OpenMP, which means it only work on shared memory systems.
//E	This is not compiled by default.

#include <iostream.h>
#include <vector.h>
#include "EMData.h"
#include "Vect.h"

#if defined(SGI)
	#include "omp.h"
#endif

using namespace std;

struct PEAK {
	float x, y, z;
	float val;
	PEAK(){x=y=z=val=0.;};
	PEAK(float tx,float ty, float tz, float tval){
		x=tx;y=ty;z=tz;val=tval;
	};
};

Vect LeastSquareFit(vector<PEAK> peaks, float thresh, float* ax, float* bx, float* ay, float* by)
{
double sx=0, sxx=0, sy=0, syy=0, sz=0, szz=0, s=0;
double tax=0, tay=0, tbx=0, tby=0;
double stt=0, w;
int i, n=0;
// ignore the weight
for(i=0;i<peaks.size();i++){
	if( peaks[i].val>=thresh) {
		n++;
		//w=sqrt(peaks[i].val);
		w=1;
		s  += w;
		sx += peaks[i].x*w;
		sy += peaks[i].y*w;
		sz += peaks[i].z*w;
	}
}
if(n<3){
	fprintf(stderr,"Error in function LeastSquareFit(): there are %d points, but at least 3 points are required for fitting\n",n);
	exit(-1);
} 
float szos=sz/s, t;
for(i=0;i<peaks.size();i++){
	if( peaks[i].val>=thresh) {
		//w=peaks[i].val;
		w=1;
		t=(peaks[i].z-szos)*w;
		stt +=t*t;
		tbx +=t*peaks[i].x*w;
		tby +=t*peaks[i].y*w;
	}
}
	
tbx/=stt;
tby/=stt;
tax=(sx-sz*tbx)/s;
tay=(sy-sz*tby)/s;

*ax=tbx;	*ay=tby;	*bx=tax;	*by=tay;
double r=sqrt(1/(tbx*tbx+tby*tby+1));
Vect ret(r*tbx,r*tby,r);
ret.normalize();
return(ret);
}

int	main(int argc, char*argv[])
{

	appinit(argc, argv);

	if (argc < 4) {
		cerr << argv[0] << ":\titerative search for the Cn symmtry axis\n";
		cerr << "Usage:	" << argv[0] << " <map> <n> <sym coeff file> [outmap=<file.mrc>] [symrad=<pixels>] [srchrad=<pixels>] [start=<x0,y0>] [zslice=<z0,z1>] [convergat=<degrees,pixels>] [thresh=<min sym coeff>] [maskout=<output map file>] [cpu=<n>] [verbose]\n";
		exit(1);
	}
	float thresh=0.75, converglimit=1*PI/180, dcenlimit=1, startx=0, starty=0;	
	int	cpu = 1, rad=0, srchrad=0, i, j, k, cn=atoi(argv[2]), z0=-1, z1=-1;
	char *symout = NULL, *maskout=NULL;
	bool verbose = false;

	if (argc > 4) {
		for (i = 4; i < argc; i++) {
			if (strncmp(argv[i], "outmap=", 7) == 0) {
				symout = &argv[i][7];
			} 
			else if (strncmp(argv[i], "maskout=", 8) == 0) {
				maskout=&argv[i][8];
			} 
			else if (strncmp(argv[i],"thresh=",7) == 0) {
				thresh=atof(&argv[i][7]);
				if(thresh<0||thresh>1){
					cerr<<argv[0]<<": \"thresh="<<thresh<<"\" is out of ranges, reset to 0.75\n";
					thresh=0.75;
				}
			}
			else if (strncmp(argv[i], "symrad=", 7) == 0) {
				rad = atoi(&argv[i][7]);
			} 
			else if (strncmp(argv[i], "srchrad=", 8) == 0) {
				srchrad = atoi(&argv[i][8]);
			} 
			else if (strncmp(argv[i], "start=", 6) == 0) {
				sscanf(&argv[i][6],"%f,%f",&startx,&starty);
			} 
			else if (strncmp(argv[i], "zslice=", 7) == 0) {
				sscanf(&argv[i][7],"%d,%d",&z0,&z1);
			} 
			else if (strncmp(argv[i], "convergat=", 10) == 0) {
				sscanf(&argv[i][10],"%f,%f",&converglimit,&dcenlimit);
				if(converglimit<0||converglimit>PI/4){
					cerr<<argv[0]<<": \"convergat="<<converglimit<<"\" is out of ranges, reset to 1 degree\n";
					converglimit=PI/180;
				}
				if(dcenlimit<0||dcenlimit>2){
					cerr<<argv[0]<<": \"convergat="<<dcenlimit<<"\" is out of ranges, reset to 1 pixel\n";
					dcenlimit=1.0;
				}
			} 
			else if (strncmp(argv[i], "cpu=", 4) == 0) {
				cpu = atoi(&argv[i][4]);
			} 
			else if (strncmp(argv[i], "verbose", 7) == 0) {
				verbose = true;
			}
			else if (strncmp(argv[i], "usefs=", 6) == 0)	;	// ignore it
			else {
				fprintf(stderr, "%s: unknown modifier '%s'\n", argv[0], argv[i]);
				exit(1);
			}
		}
	}
	
	if(cn<=1){
		cerr<<argv[0]<<": bad symmetry (Cn="<<cn<<") input\n";
		exit(2);
	}

#if defined(SGI)	
	int numprocs=omp_get_num_procs();
	if(cpu<=0)	cpu=1;
	else if (cpu>numprocs){
		cerr <<argv[0]<<": you asked "<<cpu<<" CPUs, but there are only "<<numprocs<<" CPUs for this computer "
			 <<"system, so only "<<numprocs<<" are used\n";
		cpu=numprocs;
	}
	else if (cpu==numprocs){
		cerr <<argv[0]<<": you asked "<<cpu<<" CPUs, this will take all available "<<numprocs<<" CPUs for this computer "
			 <<"system, be warned that other users might complain!\n";
	}
	else if (cpu>numprocs/2){
		cerr <<argv[0]<<": you asked "<<cpu<<" CPUs, this is more half of available "<<numprocs<<" CPUs for this computer "
			 <<"system, be warned that other users might complain!\n";
	}
	omp_set_num_threads(cpu);
	cerr.flush();
#else
	if(cpu>1){
		cerr<<argv[0]<<": this programs is parallelized with OpenMP, which is supported only on SGI at this moment, "
				<<"so the multi-processor capability is not effective on your computer now\n";
		cpu=1;
	}
#endif	
		
	EMData * map = new EMData;
	if (map->readImage(argv[1], -1)) {
		cerr << argv[0] << ": cannot read 3D map " << argv[1] << endl;
		exit(3);
	}
	int nx=map->xSize();
	int ny=map->ySize();
	int nz=map->zSize();
	if(nx!=ny || nx!=nz){
		cerr << argv[0] << ": 3D map \"" << argv[1] <<"\" is not cubic\n";
		exit(4);
	}
	
	if(rad<=0 || rad>nx/2){
		cerr << argv[0] << ": \"rad="<<rad<<"\" is out of range, changed to \""<<nx/4-1<<"\"\n";
		rad=nx/4-1;
	}
	
	if(srchrad<=0 || srchrad>=nx/2){
		cerr << argv[0] << ": \"srchrad="<<srchrad<<"\" is out of range, changed to \""<<nx/4-1<<"\"\n";
		srchrad=nx/4-1;
	}
		
	if(startx<rad-1||startx>nx-rad){
		cerr << argv[0] << ": \"startx="<<startx<<"\" is out of range, reset to " <<nx/2 <<endl;
		startx=nx/2;
	}
				
	if(starty<rad-1||starty>ny-rad){
		cerr << argv[0] << ": \"starty="<<starty<<"\" is out of range, reset to " <<ny/2 <<endl;
		starty=ny/2;
	}
	
	if(startx+srchrad+rad>=nx || startx-srchrad-rad<0) {
		cerr << argv[0] << ": \"startx="<<startx<<" rad="<<rad<<" srchrad="<<srchrad<<"\" are incompatiable" <<endl;
		exit(7);
	}
		
	if(starty+srchrad+rad>=ny || starty-srchrad-rad<0) {
		cerr << argv[0] << ": \"starty="<<starty<<" rad="<<rad<<" srchrad="<<srchrad<<"\" are incompatiable" <<endl;
		exit(8);
	}
	
	if(z0<0)	z0=0;
	if(z1<0)	z1=nz-1;
	if(z0<0 || z0>nz-1 || z0>=z1) {
		cerr << argv[0] << ": \"z0="<<z0<<"\" is out of range, reset to  0"<<endl;
		z0=0;
	}
		
	if(z1<0 || z1>nz-1 || z0>=z1) {
		cerr << argv[0] << ": \"z1="<<z1<<"\" is out of range, reset to "<<nz-1<<endl;
		z1=nz-1;
	}
		
	if(verbose) {
		cout <<argv[0]<<": start search for "<<cn<<" fold symmtry axis in 3D map "<<argv[1]<<"\n";
		cout.flush();
	}
	
	Vect oldaxis, zvect(0,0,1),currentaxis(0,0,1), realaxis(0,0,1);
	float daxis=PI/2;
	Euler rot(0,0,0);
	float oldstartx, oldstarty;
	float dcen=nx/2;
	float denthresh=map->Mean()+0*map->Sigma();	// threshold for removing the slices with little density

	EMData* d=map->copy();
	EMData* coeff=new EMData;
	coeff->setSize(nx,ny,nz);
	d->setParent(map);
	
	vector< PEAK > peaks(z1-z0+1);
	int round=0;
			
	while(daxis>converglimit || dcen>dcenlimit) {
		coeff->zero();
		oldstartx=startx;	oldstarty=starty;
			
		Euler rot2=Euler(0,2*PI/cn,0)*rot;
		
		d->setRAlign(&rot);
		//d->setTAlign(-dx,-dy,0);
		d->rotateAndTranslate(1.0,nx/2-startx,ny/2-starty,0);	// in the end, d is transformed to centered at (nx/2,ny/2) and along Z axis
				
		if(verbose){
			printf("Round %2d: rot1=%8.3f %8.3f %8.3f\trot2=%8.3f %8.3f %8.3f\tcenter=%8.3f %8.3f\n",
			round++,rot.alt()*180/PI,rot.az()*180/PI,rot.phi()*180/PI,rot2.alt()*180/PI,rot2.az()*180/PI,rot2.phi()*180/PI,startx,starty);
		}
#pragma omp parallel
		{
			EMData* d2=map->copy();
			d2->setParent(map);
#pragma omp for
			for(j=-srchrad; j<srchrad+1; j++) {
				for(i=-srchrad; i<srchrad+1; i++) {
	
					d2->setRAlign(&rot2);	// inplane rotation relative to d
					//d2->setTAlign(-dx,-dy,0);
					d2->rotateAndTranslate(1.0,nx/2-(startx+i),ny/2-(starty+j),0);
	
					float sab=0, saa=0, sbb=0, sa=0, sb=0;
					int count=0;
					for(k=z0; k<=z1; k++){
						sab=saa=sbb=sb=0;	count=0;
					    for(int jj=int(j+starty)-rad-1; jj<=int(j+starty)+rad+1; jj++){
					    	for(int ii=int(i+startx)-rad-1; ii<=int(i+startx)+rad+1; ii++){
					    		if(hypot2(ii-(startx+i),jj-(starty+j))<=rad*rad){
					    			count++;
					    			sa+=d->valueAt(ii,jj,k);
					    			sb+=d2->valueAt(ii,jj,k);
					    			saa+=d->valueAt(ii,jj,k)*d->valueAt(ii,jj,k);
					    			sbb+=d2->valueAt(ii,jj,k)*d2->valueAt(ii,jj,k);
					    			sab+=d->valueAt(ii,jj,k)*d2->valueAt(ii,jj,k);
					    		}
					    	}
					    }
						//float top=(sab-sa*sb/count);
						//float bottom=sqrt((saa-sa*sa/count)*(sbb-sb*sb/count));
						float top=sab;
						float bottom=sqrt(saa*sbb);
				   		float eff;
						if(sa/count>denthresh && sb/count>denthresh)	eff=top/bottom;
						else 			eff=0;
					    //printf("(x,y,z)=(%d,%d,%d) val=%f\ttop=%f\tbottom=%f\tcoeff=%f\n",i,j,k,eff,top,bottom,eff);
						if(eff>0 && eff<1.0) {
							coeff->setValueAt(i+nx/2,j+ny/2,k,eff);
						}
					}
				}
			}
			delete d2;
		}
		for(k=z0;k<=z1;k++){
			float maxcoeff;
			float maxx, maxy;
			maxcoeff=0;	maxx=0;	maxy=0;
			for(j=0; j<ny; j++){
				for(i=0; i<nx; i++){
					if(coeff->valueAt(i,j,k)>maxcoeff) {
						maxcoeff=coeff->valueAt(i,j,k);
						maxx=i-nx/2+startx;
						maxy=j-ny/2+starty;
					}
				}
			}
			peaks[k-z0]=PEAK(maxx,maxy,k,maxcoeff);
			if(verbose) printf("slice %3d: xc=%8.3f\tyc=%8.3f\tcoeff=%8.3f\n",k,maxx,maxy,maxcoeff);
		}
					
		float ax,bx,ay,by;
		currentaxis=LeastSquareFit(peaks, thresh, &ax, &bx, &ay, &by);
		startx=nz/2*ax+bx;
		starty=nz/2*ay+by;
		//printf("(ax=%f\tbx=%f\tay=%f\tby=%f\tstartx=%f\tstarty=%f\n",ax,bx,ay,by,startx,starty);
	
		oldaxis=realaxis;
		//realaxis=realaxis.rotate(te);
		daxis=currentaxis.angle(zvect);
		dcen=hypot(startx-oldstartx,starty-oldstarty);
		Euler te=currentaxis.euler();
		rot=te.inverse()*rot;	// accumulate rotation euler instead of successive interpolation of map
		Euler rotinv=rot.inverse();	
		realaxis=zvect.rotate(rotinv);
		if(verbose){
			printf("  Last sym axis direction=%.4f %.4f %.4f\tposition=%.4f %.4f\n",
				oldaxis.X(),oldaxis.Y(),oldaxis.Z(),oldstartx,oldstarty);
			//printf("  Current  axis direction=%.4f %.4f %.4f\tposition=%.4f %.4f\n",
			//	currentaxis.X(),currentaxis.Y(),currentaxis.Z(),startx,starty);
			printf("  New  sym axis direction=%.4f %.4f %.4f\tposition=%.4f %.4f\n",
				realaxis.X(),realaxis.Y(),realaxis.Z(),startx,starty);
			printf("  Axis   direction   diff=%.4f degrees\tcenter diff=%.4f pixels\n",
				daxis*180/PI,dcen);
		}
   	}
	
	EMData* d2=map->copy();
	coeff->zero();
	Euler rot2=Euler(0,2*PI/cn,0)*rot;
	
	// finally rotate to align the axis to Z-axis, but not centered yet	
	d->setRAlign(&rot);
	d->setTAlign(-(startx-nx/2),-(starty-ny/2),0);
	d->rotateAndTranslate();	// in the end, d is transformed to centered at (nx/2,ny/2) and along Z axis
	d2->setRAlign(&rot2);
	d2->setTAlign(-(startx-nx/2),-(starty-ny/2),0);
	d2->rotateAndTranslate();	// in the end, d2 is transformed to centered at (nx/2,ny/2) and along Z axis
	delete map;
	
	float sab=0, saa=0, sbb=0, sa=0, sb=0;
	int count=0;
	for(k=0; k<nz; k++){
		sab=saa=sbb=sb=0;	count=0;
	    for(int jj=-rad-1+ny/2; jj<=rad+1+ny/2; jj++){
	    	for(int ii=-rad-1+nx/2; ii<=rad+1+nx/2; ii++){
	    		if(hypot2(ii-nx/2,jj-ny/2)<=rad*rad){
	    			count++;
	    			sa+=d->valueAt(ii,jj,k);
	    			sb+=d2->valueAt(ii,jj,k);
	    			saa+=d->valueAt(ii,jj,k)*d->valueAt(ii,jj,k);
	    			sbb+=d2->valueAt(ii,jj,k)*d2->valueAt(ii,jj,k);
	    			sab+=d->valueAt(ii,jj,k)*d2->valueAt(ii,jj,k);
	    		}
	    	}
	    }
		//float top=(sab-sa*sb/count);
		//float bottom=sqrt((saa-sa*sa/count)*(sbb-sb*sb/count));
		float top=sab;
		float bottom=sqrt(saa*sbb);
		float eff;
		if((k>=z0 && k<=z1) || (sa/count>denthresh && sb/count>denthresh))	eff=top/bottom;
		else 			eff=0;
		if(eff>0 && eff<1.0) {
			coeff->setValueAt(nx/2,ny/2,k,eff);
		}
	}

	//output the final coeffs;
	FILE* fp=fopen(argv[3],"w");
	/*for(int i=0; i<peaks.size();i++){
		fprintf(fp,"%d\t%f\n",peaks[i].z,peaks[i].val);
	}*/
	fprintf(fp,"# Final result: rot =%8.3f %8.3f %8.3f\taxis(x,y in slice %d)=%8.3f %8.3f\n",
			rot.alt()*180/PI,rot.az()*180/PI,rot.phi()*180/PI,nz/2,startx,starty);
	for(k=0; k<nz;k++){
		fprintf(fp,"%d\t%f\n",k,coeff->valueAt(nx/2,ny/2,k));
	}
	fclose(fp);
	
	delete coeff;

	if(symout) {
 		d->writeMRC(symout);
		//d2->writeMRC("d2.mrc");
	}
	
	if(maskout){ // greedy masking out the voxels that are symmetric, instead of simple radial cut
		for(k=0;k<nx;k++){
			for(j=0;j<ny;j++){
				for(i=0;i<nx;i++){
					float dv=d->valueAt(i,j,k);
					float d2v=d2->valueAt(i,j,k);
					if((dv*dv*d2v*d2v)<1e-12) {
						d->setValueAt(i,j,k,0);
					}
					else if ((dv*d2v)/(sqrt(dv*dv*d2v*d2v))<thresh){
						d->setValueAt(i,j,k,0);
					}
				}
			}
		}
		d->writeMRC(maskout);				
	}
	delete d; delete d2;
	
	printf("Final:    rot =%8.3f %8.3f %8.3f\taxis(x,y in slice %d)=%8.3f %8.3f\n",
			rot.alt()*180/PI,rot.az()*180/PI,rot.phi()*180/PI,nz/2,startx,starty);
}


